Imports System.IO

Public Class CStatSum

  Const EPSILON = 0.0000000001
  Private Const DIGIT_MARKERS = "-+0123456789."

  Private m_bZeroX1 As Boolean
  Private m_bZeroX2 As Boolean
  Private m_bZeroY As Boolean
  Private m_bNegX1 As Boolean
  Private m_bNegX2 As Boolean
  Private m_bNegY As Boolean

  Private m_fSum As Double
  Private m_fSumX11 As Double
  Private m_fSumX21 As Double
  Private m_fSumX22 As Double
  Private m_fSumX12 As Double
  Private m_fSumY As Double
  Private m_fSumY2 As Double
  Private m_fSumX1Y As Double
  Private m_fSumX2Y As Double
  Private m_fSumX1X2 As Double
  Private m_fMeanX1 As Double
  Private m_fMeanX2 As Double
  Private m_fMeanY As Double
  Private m_fSdevX1 As Double
  Private m_fSdevX2 As Double
  Private m_fSdevY As Double
  Private m_fSlope1 As Double
  Private m_fSlope2 As Double
  Private m_fIntercept As Double
  Private m_fR2 As Double
  Private m_sTX1 As String
  Private m_sTX2 As String
  Private m_sTY As String

  Private m_nDataCount As Integer
  Private m_fX1() As Double
  Private m_fX2() As Double
  Private m_fY() As Double
  Private m_fWt() As Double


  Public Sub InitSums()
    m_fSum = 0
    m_fSumX11 = 0
    m_fSumX12 = 0
    m_fSumX21 = 0
    m_fSumX22 = 0
    m_fSumY = 0
    m_fSumY2 = 0
    m_fSumX1Y = 0
    m_fSumX2Y = 0
    m_fSumX1X2 = 0
    m_sTX1 = ""
    m_sTX2 = ""
    m_sTY = ""
  End Sub

  Public Sub New()
    InitSums()
  End Sub

  Public Function GetData(ByVal sDataFilename As String, ByRef nDataCount As Integer, _
                 ByRef fShiftArr() As Double, ByRef fScaleArr() As Double) As Boolean
    Dim sLine, sLines() As String
    Dim sData() As String
    Dim I, J, K, N As Integer
    Dim bRes As Boolean = True

    Try
      sLines = File.ReadAllLines(sDataFilename)

      nDataCount = sLines.GetUpperBound(0)

      ' Dimension arrays for maximum capacity
      ReDim m_fX1(nDataCount)
      ReDim m_fX2(nDataCount)
      ReDim m_fY(nDataCount)
      ReDim m_fWt(nDataCount)

      J = 0
      m_nDataCount = 0
      Do Until J = sLines.Length
        sLine = sLines(J).Trim()
        ' is line not empty?
        If sLine.Length > 0 Then
          ' is it NOT a comment?
          If DIGIT_MARKERS.IndexOf(sLine.Substring(0, 1)) >= 0 Then
            sData = sLine.Split(",")
            N = sData.GetUpperBound(0)
            m_fY(m_nDataCount) = Double.Parse(sData(0))
            m_fX1(m_nDataCount) = Double.Parse(sData(1))
            m_fX2(m_nDataCount) = Double.Parse(sData(1))
            If N < 3 Then
              m_fWt(m_nDataCount) = 1
            Else
              m_fWt(m_nDataCount) = Double.Parse(sData(2))
            End If
            m_nDataCount += 1
          End If
        End If
        J += 1
      Loop

      ' adjust arrays to actual number of data
      ReDim Preserve m_fX1(m_nDataCount)
      ReDim Preserve m_fX2(m_nDataCount)
      ReDim Preserve m_fY(m_nDataCount)
      ReDim Preserve m_fWt(m_nDataCount)
      nDataCount = m_nDataCount

      For I = 0 To m_nDataCount - 1
        m_fX1(I) = fScaleArr(1) * (m_fX1(I) - fShiftArr(1))
        m_fX2(I) = fScaleArr(2) * (m_fX2(I) - fShiftArr(2))
        m_fY(I) = fScaleArr(0) * (m_fY(I) - fShiftArr(0))
      Next I

    Catch ex As Exception
      bRes = False
    End Try

    Return bRes

  End Function

  Private Sub Add(ByVal X1 As Double, ByVal X2 As Double, ByVal Y As Double, Optional ByVal Wt As Double = 1)
    m_fSum = m_fSum + Wt
    m_fSumX11 = m_fSumX11 + X1 * Wt
    m_fSumX21 = m_fSumX21 + X2 * Wt
    m_fSumX12 = m_fSumX12 + X1 * X1 * Wt
    m_fSumX22 = m_fSumX22 + X2 * X2 * Wt
    m_fSumY = m_fSumY + Y * Wt
    m_fSumY2 = m_fSumY2 + Y * Y * Wt
    m_fSumX1Y = m_fSumX1Y + X1 * Y * Wt
    m_fSumX2Y = m_fSumX2Y + X2 * Y * Wt
    m_fSumX1X2 = m_fSumX1X2 + X1 * X2 * Wt
  End Sub

  Public Sub FindBestFit(ByRef objRes As CResults)
    Dim I As Integer
    Dim ITX1 As FitType
    Dim ITX2 As FitType
    Dim ITY As FitType
    Dim bOK As Boolean
    Dim fXt1 As Double
    Dim fXt2 As Double
    Dim fYt As Double
    Dim objErrs As CErrors
    Dim sModel As String
    Dim sErr As String

    objErrs = New CErrors
    Try
      m_bZeroX1 = False
      m_bZeroX1 = False
      m_bZeroY = False
      m_bNegX1 = False
      m_bNegX2 = False
      m_bNegY = False

      objRes.Clear()

      For I = 0 To m_nDataCount - 1
        If m_fX1(I) < 0 Then m_bNegX1 = True
        If m_fX2(I) < 0 Then m_bNegX2 = True
        If m_fY(I) < 0 Then m_bNegY = True
        If Math.Abs(m_fX1(I)) < EPSILON Then m_bZeroX1 = True
        If Math.Abs(m_fX2(I)) < EPSILON Then m_bZeroX2 = True
        If Math.Abs(m_fY(I)) < EPSILON Then m_bZeroY = True
      Next I

      For ITY = TYpeModule.FitType.eLinear To TYpeModule.FitType.eLn
        ' validate transformations
        If m_bZeroY And m_bNegY Then
          bOK = CanHandleZeroAndNegative(ITY)
        ElseIf m_bZeroY Then
          bOK = CanHandleZero(ITY)
        ElseIf m_bNegY Then
          bOK = CanHandleNegative(ITY)
        Else
          bOK = True
        End If

        ' Can proceed?
        If bOK Then

          For ITX1 = TYpeModule.FitType.eLinear To TYpeModule.FitType.eLn

            ' validate transformations
            If m_bZeroX1 And m_bNegX1 Then
              bOK = CanHandleZeroAndNegative(ITX1)
            ElseIf m_bZeroX1 Then
              bOK = CanHandleZero(ITX1)
            ElseIf m_bNegX1 Then
              bOK = CanHandleNegative(ITX1)
            Else
              bOK = True
            End If

            ' Can proceed?
            If bOK Then

              For ITX2 = TYpeModule.FitType.eLinear To TYpeModule.FitType.eLn

                ' validate transformations
                If m_bZeroX2 And m_bNegX2 Then
                  bOK = CanHandleZeroAndNegative(ITX2)
                ElseIf m_bZeroX2 Then
                  bOK = CanHandleZero(ITX2)
                ElseIf m_bNegX2 Then
                  bOK = CanHandleNegative(ITX2)
                Else
                  bOK = True
                End If

                ' lastly check if two transformations are the same
                If ITX1 = ITX2 Then bOK = False

                If bOK Then

                  ' initialize summations
                  InitSums()

                  For I = 0 To m_nDataCount - 1

                    Select Case ITX1
                      Case TYpeModule.FitType.eLinear
                        fXt1 = m_fX1(I)
                      Case TYpeModule.FitType.eSquare
                        fXt1 = m_fX1(I) ^ 2
                      Case TYpeModule.FitType.eCube
                        fXt1 = m_fX1(I) ^ 3
                      Case TYpeModule.FitType.eCubeRoot
                        fXt1 = m_fX1(I) ^ (1 / 3)
                      Case TYpeModule.FitType.eRecip
                        fXt1 = 1 / m_fX1(I)
                      Case TYpeModule.FitType.eRecipCubeRoot
                        fXt1 = 1 / m_fX1(I) ^ (1 / 3)
                      Case TYpeModule.FitType.eRecipSquare
                        fXt1 = 1 / m_fX1(I) ^ 2
                      Case TYpeModule.FitType.eRecipCube
                        fXt1 = 1 / m_fX1(I) ^ 3
                      Case TYpeModule.FitType.eSqrt
                        fXt1 = Math.Sqrt(m_fX1(I))
                      Case TYpeModule.FitType.eRecipSqrt
                        fXt1 = 1 / Math.Sqrt(m_fX1(I))
                      Case TYpeModule.FitType.eLn
                        fXt1 = Math.Log(m_fX1(I))
                    End Select

                    Select Case ITX2
                      Case TYpeModule.FitType.eLinear
                        fXt2 = m_fX2(I)
                      Case TYpeModule.FitType.eSquare
                        fXt2 = m_fX2(I) ^ 2
                      Case TYpeModule.FitType.eCube
                        fXt2 = m_fX2(I) ^ 3
                      Case TYpeModule.FitType.eCubeRoot
                        fXt2 = m_fX2(I) ^ (1 / 3)
                      Case TYpeModule.FitType.eRecip
                        fXt2 = 1 / m_fX2(I)
                      Case TYpeModule.FitType.eRecipCubeRoot
                        fXt2 = 1 / m_fX2(I) ^ (1 / 3)
                      Case TYpeModule.FitType.eRecipSquare
                        fXt2 = 1 / m_fX2(I) ^ 2
                      Case TYpeModule.FitType.eRecipCube
                        fXt2 = 1 / m_fX2(I) ^ 3
                      Case TYpeModule.FitType.eSqrt
                        fXt2 = Math.Sqrt(m_fX2(I))
                      Case TYpeModule.FitType.eRecipSqrt
                        fXt2 = 1 / Math.Sqrt(m_fX2(I))
                      Case TYpeModule.FitType.eLn
                        fXt2 = Math.Log(m_fX2(I))
                    End Select


                    Select Case ITY
                      Case TYpeModule.FitType.eLinear
                        fYt = m_fY(I)
                      Case TYpeModule.FitType.eSquare
                        fYt = m_fY(I) ^ 2
                      Case TYpeModule.FitType.eCube
                        fYt = m_fY(I) ^ 3
                      Case TYpeModule.FitType.eCubeRoot
                        fYt = m_fY(I) ^ (1 / 3)
                      Case TYpeModule.FitType.eRecip
                        fYt = 1 / m_fY(I)
                      Case TYpeModule.FitType.eRecipCubeRoot
                        fYt = 1 / m_fY(I) ^ (1 / 3)
                      Case TYpeModule.FitType.eRecipSquare
                        fYt = 1 / m_fY(I) ^ 2
                      Case TYpeModule.FitType.eRecipCube
                        fYt = 1 / m_fY(I) ^ 3
                      Case TYpeModule.FitType.eSqrt
                        fYt = Math.Sqrt(m_fY(I))
                      Case TYpeModule.FitType.eRecipSqrt
                        fYt = 1 / Math.Sqrt(m_fY(I))
                      Case TYpeModule.FitType.eLn
                        fYt = Math.Log(m_fY(I))
                    End Select

                    ' add transformed data to statistical summations
                    Add(fXt1, fXt2, fYt, m_fWt(I))

                  Next I

                  ' store transformation data
                  m_sTX1 = SayTransform(ITX1, "X1")
                  m_sTX2 = SayTransform(ITX2, "X2")
                  m_sTY = SayTransform(ITY, "Y")
                  sModel = m_sTY & " = A + B1 * " & m_sTX1 & " + B2 * " & m_sTX2
                  ' calculate regression statistics and store in
                  ' object accessed by m_objRes
                  CalcLR(objRes, objErrs)

                  If objErrs.GetCount > 0 Then
                    sErr = objErrs.GetErrText(0)
                  Else
                    sErr = ""
                  End If
                  objErrs.Clear() ' reset error object

                  objRes.Add(sModel, m_fR2, m_fSlope1, m_fSlope2, m_fIntercept, sErr)

                End If

              Next ITX2

            End If

          Next ITX1

        End If
      Next ITY

    Catch ex As Exception
      objErrs.Add(ex.Message)
    End Try

  End Sub

  Private Function SayTransform(ByVal eVal As FitType, Optional ByVal sVar As String = "X") As String
    Select Case eVal
      Case TYpeModule.FitType.eLinear
        Return sVar
      Case TYpeModule.FitType.eSquare
        Return sVar & "^2"
      Case TYpeModule.FitType.eCube
        Return sVar & "^3"
      Case TYpeModule.FitType.eCubeRoot
        Return sVar & "^1/3"
      Case TYpeModule.FitType.eRecip
        Return "1/" & sVar
      Case TYpeModule.FitType.eRecipCubeRoot
        Return "1/" & sVar & "^1/3"
      Case TYpeModule.FitType.eRecipSquare
        Return "1/" & sVar & "^2"
      Case TYpeModule.FitType.eRecipCube
        Return "1/" & sVar & "^3"
      Case TYpeModule.FitType.eSqrt
        Return sVar & "^1/2"
      Case TYpeModule.FitType.eRecipSqrt
        Return "1/" & sVar & "^1/2"
      Case TYpeModule.FitType.eLn
        Return "Ln(" & sVar & ")"
    End Select
  End Function

  Private Function CanHandleZero(ByVal eVal As FitType) As Boolean

    Select Case eVal
      Case TYpeModule.FitType.eLinear, TYpeModule.FitType.eSquare, _
           TYpeModule.FitType.eCube, TYpeModule.FitType.eCubeRoot, TYpeModule.FitType.eSqrt
        Return True
      Case Else
        Return False
    End Select
  End Function

  Private Function CanHandleZeroAndNegative(ByVal eVal As FitType) As Boolean

    Select Case eVal
      Case TYpeModule.FitType.eLinear, TYpeModule.FitType.eSquare, TYpeModule.FitType.eCube
        Return True
      Case Else
        Return False
    End Select
  End Function

  Private Function CanHandleNegative(ByVal eVal As FitType) As Boolean

    Select Case eVal
      Case TYpeModule.FitType.eLinear, TYpeModule.FitType.eSquare, TYpeModule.FitType.eCube, _
          TYpeModule.FitType.eRecip, TYpeModule.FitType.eRecipSquare, TYpeModule.FitType.eRecipCube
        Return True
      Case Else
        Return False
    End Select
  End Function


  Private Sub CalcLR(ByRef objRes As CResults, ByRef objErrs As CErrors)
    Dim A As Double
    Dim B As Double

    If m_fSum < 2 Then Exit Sub
    ' caluclate regression
    Try
      m_fMeanX1 = m_fSumX11 / m_fSum
      m_fMeanX2 = m_fSumX21 / m_fSum
      m_fMeanY = m_fSumY / m_fSum
      m_fSdevX1 = Math.Sqrt((m_fSumX12 - m_fSumX11 ^ 2 / m_fSum) / (m_fSum - 1))
      m_fSdevX2 = Math.Sqrt((m_fSumX22 - m_fSumX21 ^ 2 / m_fSum) / (m_fSum - 1))
      m_fSdevY = Math.Sqrt((m_fSumY2 - m_fSumY ^ 2 / m_fSum) / (m_fSum - 1))

      A = (m_fSum * m_fSumX12 - m_fSumX11 ^ 2) * _
          (m_fSum * m_fSumX2Y - m_fSumX21 * m_fSumY)
      B = (m_fSum * m_fSumX1X2 - m_fSumX11 * m_fSumX21) * _
          (m_fSum * m_fSumX1Y - m_fSumX11 * m_fSumY)

      m_fSlope2 = (A - B) / _
         ((m_fSum * m_fSumX12 - m_fSumX11 ^ 2) * _
         (m_fSum * m_fSumX22 - m_fSumX21 ^ 2) - _
         (m_fSum * m_fSumX1X2 - m_fSumX11 * m_fSumX21) ^ 2)
      m_fSlope1 = ((m_fSum * m_fSumX1Y - m_fSumX11 * m_fSumY) - _
        m_fSlope2 * (m_fSum * m_fSumX1X2 - m_fSumX11 * m_fSumX21)) / _
        (m_fSum * m_fSumX12 - m_fSumX11 ^ 2)
      m_fIntercept = m_fMeanY - m_fSlope1 * m_fMeanX1 - m_fSlope2 * m_fMeanX2
      m_fR2 = (m_fIntercept * m_fSumY + m_fSlope1 * m_fSumX1Y + m_fSlope2 * m_fSumX2Y - m_fSumY ^ 2 / m_fSum) / _
              (m_fSumY2 - m_fSumY ^ 2 / m_fSum)

    Catch ex As Exception
      objErrs.Add("Error in model " & m_sTY = "A0 + A1 " & _
                   m_sTX1 & " + A2 " & m_sTX2 & vbCrLf & ex.Message)
    End Try
  End Sub

End Class
